arXiv is now an independent nonprofit! Learn more
License: CC BY-NC-ND 4.0
arXiv:2605.00837v1 [cs.LG] 04 Apr 2026

Fast Log-Domain Sinkhorn Optimal Transport
with Warp-Level GPU Reductions

Hao Xiao
ATLAS AI Lab
xiaohao@atlasthinktank.com
Abstract

Entropic regularized optimal transport (OT) via the Sinkhorn algorithm has become a fundamental tool in machine learning, yet existing implementations either suffer from numerical instability for small regularization parameters or incur significant overhead from deep learning frameworks. We present FastSinkhorn, a lightweight, native CUDA implementation of the log-domain Sinkhorn algorithm that combines warp-level shuffle reductions with shared-memory tiling to achieve high GPU utilization without sacrificing numerical stability. Our solver operates entirely in the log-domain, enabling robust computation for regularization parameters as small as ε=104\varepsilon=10^{-4} where standard-domain methods fail. On dense OT problems with n=m=8192n=m=8192, our implementation achieves 12×12\times speedup over the widely-used POT library and 5.9×5.9\times speedup over GPU-accelerated PyTorch baselines, while consuming only 256 MB of GPU memory. We validate our solver on image color transfer, 3D point cloud matching, and convergence analysis, demonstrating that native CUDA kernels with careful numerical treatment provide a practical and efficient foundation for large-scale optimal transport computation.

1 Introduction

Optimal transport (OT) provides a mathematically principled framework for comparing probability distributions by computing the minimum-cost coupling between them (Villani, 2003, 2009). In the discrete setting, the Kantorovich formulation reduces to a linear program over coupling matrices, with applications spanning generative modeling (Arjovsky et al., 2017), domain adaptation (Courty et al., 2017), document similarity (Kusner et al., 2015), and geometric processing (Solomon et al., 2015).

The computational bottleneck of exact OT solvers—scaling as O(n3logn)O(n^{3}\log n) via the network simplex algorithm—was dramatically alleviated by Cuturi (2013), who introduced entropic regularization and showed that the resulting problem can be solved via Sinkhorn’s matrix scaling algorithm in near-linear time per iteration. This breakthrough has led to widespread adoption of the Sinkhorn algorithm in machine learning pipelines (Peyré and Cuturi, 2019).

Despite its popularity, practical deployment of the Sinkhorn algorithm on modern hardware faces two key challenges:

Numerical instability.

The standard-domain Sinkhorn algorithm computes the Gibbs kernel Kij=exp(Cij/ε)K_{ij}=\exp(-C_{ij}/\varepsilon), which causes catastrophic overflow or underflow in floating-point arithmetic when the regularization parameter ε\varepsilon is small—precisely the regime needed for accurate OT approximation. Log-domain stabilization (Schmitzer, 2019) addresses this but has not been widely adopted in GPU-optimized implementations.

Framework overhead.

Existing GPU implementations rely on deep learning frameworks (PyTorch, JAX) or specialized libraries (KeOps/GeomLoss (Feydy et al., 2019; Charlier et al., 2021)), which introduce significant overhead from automatic differentiation graphs, memory allocators, and Python-to-GPU dispatch. For applications that require only the forward OT computation—such as point cloud registration, color transfer, or batch distance computation—this overhead is unnecessary.

1.1 Contributions

We present FastSinkhorn, a lightweight CUDA C++ library for entropic regularized optimal transport. Our contributions are:

  1. 1.

    Numerically stable GPU solver. A log-domain Sinkhorn implementation with a two-pass LogSumExp reduction that operates entirely in log-space, handling regularization parameters as small as ε=104\varepsilon=10^{-4} without numerical failure.

  2. 2.

    Warp-level GPU optimizations. We exploit CUDA warp shuffle instructions (__shfl_down_sync) for intra-warp reductions, combined with shared memory for cross-warp communication. This hierarchical reduction eliminates shared-memory bank conflicts within warps and reduces synchronization overhead by 1.93×1.93\times compared to pure shared-memory reductions.

  3. 3.

    Comprehensive evaluation. We provide extensive benchmarks against POT (Flamary et al., 2021), GeomLoss (Feydy et al., 2019), and PyTorch-based solvers, along with ablation studies quantifying the contribution of each optimization, numerical stability analysis, and real-world applications in image color transfer and point cloud matching.

Our solver is self-contained (no dependencies beyond the CUDA runtime), supports both pre-computed cost matrices and Euclidean point cloud inputs, and is publicly available under the MIT license.

2 Related Work

Entropic optimal transport.

The use of entropic regularization for optimal transport was popularized by Cuturi (2013), who showed that adding a KL-divergence penalty transforms the linear program into a strictly convex problem solvable via Sinkhorn iterations (Sinkhorn, 1967). The comprehensive treatment by Peyré and Cuturi (2019) establishes the theoretical foundations, including convergence rates and approximation guarantees. Altschuler et al. (2017) provided near-linear time complexity bounds, showing that O~(n2/ε2)\tilde{O}(n^{2}/\varepsilon^{2}) operations suffice for an ε\varepsilon-approximate solution.

Numerical stabilization.

The standard Sinkhorn algorithm suffers from numerical overflow/underflow when ε\varepsilon is small, as the Gibbs kernel entries Kij=eCij/εK_{ij}=e^{-C_{ij}/\varepsilon} span extreme ranges. Schmitzer (2019) proposed log-domain stabilization, and Chizat et al. (2018) extended this to unbalanced settings. While these stabilizations are well-understood theoretically, many GPU implementations still default to the numerically fragile standard-domain formulation.

GPU-accelerated solvers.

Several GPU-based OT solvers exist in the literature. The POT library (Flamary et al., 2021) provides a comprehensive Python API with optional GPU acceleration via CuPy. GeomLoss (Feydy et al., 2019) and its KeOps backend (Charlier et al., 2021) offer GPU-accelerated Sinkhorn divergences with automatic differentiation, designed for integration with PyTorch. These framework-based approaches provide flexibility but carry overhead from the Python runtime, autograd engine, and generic GPU memory management.

Algorithmic acceleration.

Beyond GPU parallelization, several algorithmic improvements reduce the computational cost of OT. Multiscale approaches (Gerber and Maggioni, 2017; Schmitzer, 2019) use coarse-to-fine strategies with sparse representations. Low-rank factorizations (Scetbon et al., 2021) approximate the transport plan with reduced-rank factors, achieving sub-quadratic complexity. Accelerated gradient methods (Dvurechensky et al., 2018; Lin et al., 2019) provide improved convergence rates for the dual problem. These approaches are complementary to our GPU optimization strategy and could be combined in future work.

Our position.

Our work occupies a distinct niche: a native CUDA implementation that combines log-domain stability with warp-level GPU optimizations, targeting the common case of dense, moderate-sized (n,m16384n,m\leq 16384) OT problems where the full cost matrix fits in GPU memory. Unlike framework-based approaches, our solver has zero Python overhead and minimal memory footprint, making it suitable for latency-sensitive applications and embedding in C++ pipelines.

3 Background

3.1 Discrete Optimal Transport

Let 𝝁=(μ1,,μn)Σn\bm{\mu}=(\mu_{1},\ldots,\mu_{n})\in\Sigma_{n} and 𝝂=(ν1,,νm)Σm\bm{\nu}=(\nu_{1},\ldots,\nu_{m})\in\Sigma_{m} be discrete probability distributions on point sets {xi}i=1n\{x_{i}\}_{i=1}^{n} and {yj}j=1m\{y_{j}\}_{j=1}^{m}, where Σk={a+k:iai=1}\Sigma_{k}=\{a\in\mathbb{R}_{+}^{k}:\sum_{i}a_{i}=1\} denotes the probability simplex. Given a cost matrix C+n×mC\in\mathbb{R}_{+}^{n\times m} with Cij=c(xi,yj)C_{ij}=c(x_{i},y_{j}), the Kantorovich optimal transport problem seeks the minimum-cost coupling:

𝒯c(𝝁,𝝂)=min𝝅Π(𝝁,𝝂)C,𝝅F=min𝝅Π(𝝁,𝝂)i,jCijπij,\mathcal{T}_{c}(\bm{\mu},\bm{\nu})=\min_{\bm{\pi}\in\Pi(\bm{\mu},\bm{\nu})}\langle C,\bm{\pi}\rangle_{F}=\min_{\bm{\pi}\in\Pi(\bm{\mu},\bm{\nu})}\sum_{i,j}C_{ij}\pi_{ij}, (1)

where the feasible set of couplings (transport plans) is:

Π(𝝁,𝝂)={𝝅+n×m|𝝅𝟏m=𝝁,𝝅𝟏n=𝝂}.\Pi(\bm{\mu},\bm{\nu})=\left\{\bm{\pi}\in\mathbb{R}_{+}^{n\times m}\;\middle|\;\bm{\pi}\mathbf{1}_{m}=\bm{\mu},\quad\bm{\pi}^{\top}\mathbf{1}_{n}=\bm{\nu}\right\}. (2)

This is a linear program in nmnm variables with n+mn+m equality constraints. Exact solvers based on the network simplex algorithm have O(n3logn)O(n^{3}\log n) complexity, which is prohibitive for large nn (Peyré and Cuturi, 2019).

3.2 Entropic Regularization

Following Cuturi (2013), we consider the entropically regularized problem:

𝒯cε(𝝁,𝝂)=min𝝅Π(𝝁,𝝂)C,𝝅+εKL(𝝅𝝁𝝂),\mathcal{T}_{c}^{\varepsilon}(\bm{\mu},\bm{\nu})=\min_{\bm{\pi}\in\Pi(\bm{\mu},\bm{\nu})}\langle C,\bm{\pi}\rangle+\varepsilon\,\mathrm{KL}(\bm{\pi}\|\bm{\mu}\otimes\bm{\nu}), (3)

where ε>0\varepsilon>0 is the regularization parameter and the KL divergence is:

KL(𝝅𝝁𝝂)=i,jπij(logπijμiνj1)+1.\mathrm{KL}(\bm{\pi}\|\bm{\mu}\otimes\bm{\nu})=\sum_{i,j}\pi_{ij}\left(\log\frac{\pi_{ij}}{\mu_{i}\nu_{j}}-1\right)+1. (4)

The addition of the strictly convex KL term ensures a unique minimizer 𝝅ε\bm{\pi}^{\varepsilon}. As ε0+\varepsilon\to 0^{+}, we recover the solution to the original problem: 𝝅ε𝝅\bm{\pi}^{\varepsilon}\to\bm{\pi}^{*} (Peyré and Cuturi, 2019).

3.3 Structure of the Optimal Solution

Writing the KKT conditions with dual variables 𝜶n\bm{\alpha}\in\mathbb{R}^{n} (for row constraints) and 𝜷m\bm{\beta}\in\mathbb{R}^{m} (for column constraints), and setting /πij=0\partial\mathcal{L}/\partial\pi_{ij}=0, the optimal transport plan has the Gibbs kernel structure:

πijε=μiνjexp(αi+βjCijε).\pi_{ij}^{\varepsilon}=\mu_{i}\,\nu_{j}\,\exp\!\left(\frac{\alpha_{i}+\beta_{j}-C_{ij}}{\varepsilon}\right). (5)

3.4 The Sinkhorn Algorithm

Substituting Eq. (5) into the marginal constraints yields a fixed-point iteration. The Sinkhorn algorithm alternates:

αi(k+1)\displaystyle\alpha_{i}^{(k+1)} =εlogj=1mνjexp(βj(k)Cijε),\displaystyle=-\varepsilon\log\sum_{j=1}^{m}\nu_{j}\exp\!\left(\frac{\beta_{j}^{(k)}-C_{ij}}{\varepsilon}\right), (6)
βj(k+1)\displaystyle\beta_{j}^{(k+1)} =εlogi=1nμiexp(αi(k+1)Cijε),\displaystyle=-\varepsilon\log\sum_{i=1}^{n}\mu_{i}\exp\!\left(\frac{\alpha_{i}^{(k+1)}-C_{ij}}{\varepsilon}\right), (7)

starting from 𝜶(0)=𝟎\bm{\alpha}^{(0)}=\mathbf{0}, 𝜷(0)=𝟎\bm{\beta}^{(0)}=\mathbf{0}.

Theorem 1 (Linear convergence (Peyré and Cuturi, 2019, Theorem 4.2)).

The Sinkhorn iterates converge linearly: 𝛂(k)𝛂λk𝛂(0)𝛂\|\bm{\alpha}^{(k)}-\bm{\alpha}^{*}\|_{\infty}\leq\lambda^{k}\|\bm{\alpha}^{(0)}-\bm{\alpha}^{*}\|_{\infty}, where λ=e2R/ε\lambda=e^{-2R/\varepsilon} and R=maxi,jCijmini,jCijR=\max_{i,j}C_{ij}-\min_{i,j}C_{ij} is the cost range.

The contraction rate λ\lambda approaches 1 as ε0\varepsilon\to 0, revealing the fundamental tension: smaller ε\varepsilon yields a better OT approximation but requires more iterations.

4 Method: GPU-Optimized Log-Domain Sinkhorn

4.1 Log-Domain Stabilization

The standard Sinkhorn algorithm (Section 3) computes the Gibbs kernel Kij=exp(Cij/ε)K_{ij}=\exp(-C_{ij}/\varepsilon) explicitly. For typical cost matrices with Cij[0,10]C_{ij}\in[0,10] and ε=0.01\varepsilon=0.01, the kernel entries span [e1000,1][e^{-1000},1]—far beyond the range of IEEE 754 single-precision floats ([1038,1038]\approx[10^{-38},10^{38}]). This causes catastrophic underflow.

We instead work entirely in the log-domain. The key operation is the numerically stable LogSumExp:

LSE(x1,,xm)=maxjxj+logj=1mexp(xjmaxjxj).\mathrm{LSE}(x_{1},\ldots,x_{m})=\max_{j}x_{j}+\log\sum_{j=1}^{m}\exp(x_{j}-\max_{j}x_{j}). (8)

By subtracting the maximum before exponentiation, the largest exponential becomes e0=1e^{0}=1 (no overflow) and all others satisfy exjM1e^{x_{j}-M}\leq 1 (no overflow). The sum is at least 1, so the logarithm is non-negative (no underflow issue).

The Sinkhorn dual updates (Eqs. 67) become:

αi\displaystyle\alpha_{i} =εLSEj(βjCijε+logνj),\displaystyle=-\varepsilon\cdot\mathrm{LSE}_{j}\!\left(\frac{\beta_{j}-C_{ij}}{\varepsilon}+\log\nu_{j}\right), (9)
βj\displaystyle\beta_{j} =εLSEi(αiCijε+logμi).\displaystyle=-\varepsilon\cdot\mathrm{LSE}_{i}\!\left(\frac{\alpha_{i}-C_{ij}}{\varepsilon}+\log\mu_{i}\right). (10)

4.2 CUDA Parallelization Strategy

Each dual update requires computing one LogSumExp per row (for 𝜶\bm{\alpha}) or per column (for 𝜷\bm{\beta}). We assign one CUDA thread block per row/column, with threads within the block cooperating on the reduction. This provides nn (or mm) independent blocks—sufficient to saturate modern GPUs for n256n\geq 256.

Algorithm 1 presents the complete procedure.

Algorithm 1 Log-Domain Sinkhorn with CUDA Parallelization
0: Cost matrix Cn×mC\in\mathbb{R}^{n\times m}, distributions 𝝁Σn\bm{\mu}\in\Sigma_{n}, 𝝂Σm\bm{\nu}\in\Sigma_{m}, regularization ε>0\varepsilon>0, tolerance τ\tau, max iterations KK, check interval cc
0: Transport cost C,𝝅ε\langle C,\bm{\pi}^{\varepsilon}\rangle, dual potentials (𝜶,𝜷)(\bm{\alpha},\bm{\beta})
1: Allocate GPU: 𝐝_𝐂[n×m]\mathbf{d\_C}[n\times m], 𝐝_α[n]𝟎\mathbf{d\_\alpha}[n]\leftarrow\mathbf{0}, 𝐝_β[m]𝟎\mathbf{d\_\beta}[m]\leftarrow\mathbf{0}
2: Precompute: 𝐝_𝐥𝐨𝐠_μ[i]logμi\mathbf{d\_log\_\mu}[i]\leftarrow\log\mu_{i}, 𝐝_𝐥𝐨𝐠_ν[j]logνj\mathbf{d\_log\_\nu}[j]\leftarrow\log\nu_{j}
3:for k=1,2,,Kk=1,2,\ldots,K do
4:  updateAlphaKerneln,B\langle\langle\langle n,B\rangle\rangle\rangle: {one block per row}
5:   αiεBlockLSEj(βjCijε+logνj)\alpha_{i}\leftarrow-\varepsilon\cdot\mathrm{BlockLSE}_{j}\!\left(\frac{\beta_{j}-C_{ij}}{\varepsilon}+\log\nu_{j}\right) for each block ii
6:  updateBetaKernelm,B\langle\langle\langle m,B\rangle\rangle\rangle: {one block per column}
7:   βjεBlockLSEi(αiCijε+logμi)\beta_{j}\leftarrow-\varepsilon\cdot\mathrm{BlockLSE}_{i}\!\left(\frac{\alpha_{i}-C_{ij}}{\varepsilon}+\log\mu_{i}\right) for each block jj
8:  if kmodc=0k\bmod c=0 then
9:   Compute marginal error: err=i|riμi|\mathrm{err}=\sum_{i}|r_{i}-\mu_{i}| {amortize sync}
10:   if err<τ\mathrm{err}<\tau then
11:    break (converged)
12:   end if
13:  end if
14:end for
15: Compute transport cost: C,𝝅ε=i,jCijexp(αi+βjCijε+logμi+logνj)\langle C,\bm{\pi}^{\varepsilon}\rangle=\sum_{i,j}C_{ij}\exp\!\left(\frac{\alpha_{i}+\beta_{j}-C_{ij}}{\varepsilon}+\log\mu_{i}+\log\nu_{j}\right)
16:return transport cost, (𝜶,𝜷)(\bm{\alpha},\bm{\beta})

4.3 Warp-Level Reductions

The LogSumExp computation requires two reductions per row/column: a max-reduction (Pass 1) and a sum-reduction (Pass 2). We implement these using a hierarchical reduction scheme:

Intra-warp reduction.

Within each warp (32 threads), we use the __shfl_down_sync instruction to perform butterfly reductions in log2(32)=5\log_{2}(32)=5 steps, without any shared memory access or synchronization barriers. For the max-reduction:

valmax(val,__shfl_down_sync(0xFFFFFFFF,val,offset))\texttt{val}\leftarrow\max(\texttt{val},\ \texttt{\_\_shfl\_down\_sync}(\texttt{0xFFFFFFFF},\texttt{val},\texttt{offset})) (11)

for offsets 16,8,4,2,116,8,4,2,1. The sum-reduction replaces max\max with ++.

Cross-warp reduction.

With block size B=256B=256 (8 warps), each warp’s result is written to shared memory (one slot per warp). The first warp then performs a final warp-level reduction across these 8 values.

This two-level design eliminates shared memory bank conflicts for intra-warp communication, avoids __syncthreads() barriers within warps, and requires only 3232 shared memory slots regardless of block size.

4.4 Memory Layout and Access Patterns

The cost matrix CC dominates memory usage at O(nm)O(nm) floats, while dual potentials require only O(n+m)O(n+m)—negligible in comparison. For n=m=8192n=m=8192, the cost matrix occupies 256 MB and the potentials add \sim64 KB.

Coalesced access.

The cost matrix is stored in row-major order. In the 𝜶\bm{\alpha}-update kernel, block ii reads row ii of CC sequentially—threads within a warp access consecutive addresses, achieving coalesced global memory loads. The 𝜷\bm{\beta}-update kernel accesses column jj with stride mm, which is less favorable; however, the L2 cache (typically 4–6 MB on modern GPUs) mitigates this when the cost matrix fits in cache.

Precomputed log-weights.

We precompute logμi\log\mu_{i} and logνj\log\nu_{j} on the host and transfer them once, avoiding repeated logarithm computations on the GPU.

4.5 Convergence Monitoring

The marginal error err=i|riμi|\mathrm{err}=\sum_{i}|r_{i}-\mu_{i}|, where ri=jπijεr_{i}=\sum_{j}\pi^{\varepsilon}_{ij} is the row marginal, requires a GPU-to-CPU data transfer for the scalar result. To amortize this synchronization cost, we check convergence every c=10c=10 iterations (configurable). The row marginal is computed in log-domain:

logri=logμi+LSEj(αi+βjCijε+logνj),\log r_{i}=\log\mu_{i}+\mathrm{LSE}_{j}\!\left(\frac{\alpha_{i}+\beta_{j}-C_{ij}}{\varepsilon}+\log\nu_{j}\right), (12)

which reuses the same LogSumExp kernel structure as the dual updates.

4.6 Complexity Analysis

Each Sinkhorn iteration performs two LogSumExp reductions over the n×mn\times m cost matrix, giving O(nm)O(nm) work per iteration. With KK iterations and convergence checks every cc iterations, the total complexity is O(Knm+Kcnm)O(Knm+\frac{K}{c}nm). Since cc is a small constant, this simplifies to O(Knm)O(Knm). On the GPU, the nn (or mm) independent reductions execute in parallel across blocks, achieving an effective wall-clock complexity of O(Km)O(Km) (or O(Kn)O(Kn)) per SM, distributed across the available streaming multiprocessors.

5 Experiments

We evaluate FastSinkhorn across six experimental dimensions: baseline comparison, scaling behavior, ablation study, numerical stability, convergence analysis, and real-world applications. Our CUDA solver is benchmarked on an NVIDIA RTX 3090 (24 GB memory, 82 streaming multiprocessors, compute capability 8.6). Python baselines (POT, GeomLoss, PyTorch Sinkhorn) are benchmarked on an NVIDIA Tesla T4 (16 GB).

5.1 Baseline Comparisons

We compare against three widely-used implementations:

  • POT (Flamary et al., 2021): Python Optimal Transport library, CPU backend (NumPy).

  • GeomLoss (Feydy et al., 2019): PyTorch-based Sinkhorn with KeOps GPU backend.

  • PyTorch Sinkhorn: Log-domain Sinkhorn implemented in pure PyTorch (GPU).

All methods solve the same problem: OT between two random distributions on a uniform 1D grid with squared Euclidean cost and ε=0.01\varepsilon=0.01. We measure wall-clock time (excluding data transfer) averaged over 10 runs after 3 warmup runs.

Table 1: Wall-clock time (ms) for varying problem sizes with ε=0.01\varepsilon=0.01. Our solver (RTX 3090) vs. baselines (POT on CPU; GeomLoss and PyTorch on Tesla T4).
Method n=256n{=}256 n=512n{=}512 n=1024n{=}1024 n=2048n{=}2048 n=4096n{=}4096 n=8192n{=}8192
POT (CPU) 3.2 13.1 55.1 256.0 1097.2 4462.1
GeomLoss (T4) 37.1 24.1 23.5 80.0 304.8 1193.5
PyTorch Sinkhorn (T4) 10.5 10.3 36.6 135.9 533.5 2174.5
Ours (RTX 3090) 0.3 1.3 5.1 20.7 89.1 371.6
Speedup vs. POT 10.0×\times 10.4×\times 10.8×\times 12.4×\times 12.3×\times 12.0×\times
Speedup vs. PyTorch 32.4×\times 8.2×\times 7.2×\times 6.6×\times 6.0×\times 5.9×\times

Table 1 and Figure 1 show the results. Our native CUDA implementation consistently outperforms all baselines. Compared to POT (CPU), we achieve 101012×12\times speedup, with the gap widening at larger problem sizes. Compared to PyTorch Sinkhorn on a Tesla T4, we achieve 5.95.932×32\times speedup, demonstrating the benefit of eliminating framework overhead. Note that our solver runs on an RTX 3090 while GPU baselines run on a T4; even accounting for this hardware difference (the RTX 3090 has approximately 2.5×2.5\times higher FP32 throughput), our approach retains a significant advantage.

Refer to caption
Figure 1: Wall-clock time vs. problem size n=mn=m on log-log scale (ε=0.01\varepsilon=0.01). Our solver scales as O(n2)O(n^{2}) with a smaller constant than framework-based approaches.

5.2 Scaling Behavior

We evaluate our solver for problem sizes from n=64n=64 to n=16,384n=16{,}384 with ε=0.01\varepsilon=0.01. Figure 2 shows the results. Computation time scales quadratically with nn, consistent with the O(n2)O(n^{2}) per-iteration cost of dense matrix operations. Peak GPU memory is dominated by the n×nn\times n cost matrix (e.g., 1024 MB for n=16,384n=16{,}384). The iteration count remains approximately constant (\sim180–200) across problem sizes, confirming that the convergence rate of Sinkhorn’s algorithm depends on ε\varepsilon rather than nn (Theorem 1). Our solver handles n=16,384n=16{,}384 in 1.54 seconds on a single RTX 3090.

Refer to caption
Figure 2: Scaling analysis of FastSinkhorn (ε=0.01\varepsilon=0.01). (a) Computation time scales quadratically with nn. (b) Peak GPU memory is dominated by the n2n^{2} cost matrix. (c) Iteration count is approximately independent of nn for fixed ε\varepsilon.

5.3 Ablation Study

We ablate each optimization to quantify its individual contribution, using n=m=2048n=m=2048 and ε=0.01\varepsilon=0.01 as the reference configuration.

Table 2: Ablation study: time (ms) for n=m=2048n=m=2048, ε=0.01\varepsilon=0.01. Each row removes one optimization from the full system.
Configuration Time (ms) Slowdown
Full system (ours) 14.8 1.00×1.00\times
Shared-memory only (no warp shuffle) 28.5 1.93×1.93\times
Standard-domain (no log-LSE) 16.9 1.14×1.14\times
Block size = 64 (vs. 256) 21.1 1.42×1.42\times
Block size = 128 16.7 1.13×1.13\times
Block size = 512 16.8 1.14×1.14\times
Convergence check every iteration 24.8 1.68×1.68\times
Check interval = 5 16.8 1.14×1.14\times
Check interval = 10 15.2 1.03×1.03\times

Standard-domain solver fails (NaN) for ε<0.005\varepsilon<0.005; time reported for ε=0.01\varepsilon=0.01 only.

Table 2 reveals that warp-level shuffle reductions provide the single largest speedup (1.93×1.93\times), followed by the convergence check interval optimization (1.68×1.68\times when checking every iteration vs. every 20 iterations). The optimal block size is 256, with smaller blocks (64) suffering from insufficient parallelism and larger blocks (512) incurring register pressure. The log-domain formulation adds only 14%14\% overhead compared to the standard-domain variant, a modest price for the substantially improved numerical stability (Section 5.4).

Refer to caption
Figure 3: Ablation study: contribution of each optimization to overall performance. Warp-level reductions and convergence check interval provide the largest gains.

5.4 Numerical Stability

A critical advantage of our log-domain formulation is robustness for small ε\varepsilon. We sweep ε\varepsilon from 1.01.0 to 10410^{-4} with n=512n=512 and compare:

  • Log-domain (ours): computes dual updates via LogSumExp (Eqs. 910).

  • Standard-domain: computes Gibbs kernel K=exp(C/ε)K=\exp(-C/\varepsilon) and scaling vectors u,vu,v.

Figure 4 shows the results. The standard-domain solver produces NaN for ε<0.005\varepsilon<0.005 due to overflow in the Gibbs kernel exp(Cij/ε)\exp(-C_{ij}/\varepsilon). In contrast, our log-domain solver remains numerically stable and converges successfully for all tested values down to ε=104\varepsilon=10^{-4}. Both methods produce comparable transport costs for ε0.01\varepsilon\geq 0.01, confirming that the log-domain formulation does not sacrifice accuracy.

Refer to caption
Figure 4: Numerical stability comparison (n=512n=512). (a) Transport cost as a function of ε\varepsilon. The standard-domain solver produces NaN for ε<0.005\varepsilon<0.005, while our log-domain solver remains stable down to ε=104\varepsilon=10^{-4}. (b) Convergence status across (ε,maxC)(\varepsilon,\max C) configurations: converged (green), diverged (yellow), NaN (red).

5.5 Convergence Analysis

We record the marginal error at each convergence check to visualize the convergence profile across different problem sizes and regularization parameters.

Refer to caption
Figure 5: Convergence profiles for different ε\varepsilon values and problem sizes. All configurations converge to machine-precision marginal error within a small number of iterations, with smaller ε\varepsilon requiring more iterations.

5.6 Applications

We demonstrate practical utility on two tasks:

Image color transfer.

Given a source and target image, we sample N=512N=512 pixels from each, compute the OT plan between their RGB distributions (ε=0.01\varepsilon=0.01), and apply barycentric mapping to transfer the color palette. Figure 6 shows qualitative results: the warm color palette of the source image is successfully transferred to match the cool target palette.

3D point cloud matching.

We compute OT correspondences between two 3D point clouds (n=200n=200 points each), where the target is a rotated and translated copy of the source with added Gaussian noise. Figure 7 shows that the OT plan recovers accurate correspondences between the two clouds.

Refer to caption
Figure 6: Image color transfer via optimal transport. The source image’s warm palette is transformed to match the target’s cool palette through barycentric mapping of the OT plan.
Refer to caption
Figure 7: 3D point cloud matching. (a) Source and target point clouds. (b) OT correspondences (green lines) connecting matched points.

6 Conclusion

We presented FastSinkhorn, a native CUDA implementation of the log-domain Sinkhorn algorithm for entropic regularized optimal transport. By combining numerically stable LogSumExp computation with warp-level shuffle reductions and shared-memory tiling, our solver achieves 12×12\times speedup over the POT library and 5.9×5.9\times over PyTorch-based GPU solvers on dense OT problems up to n=m=8192n=m=8192. The log-domain formulation enables robust computation for regularization parameters as small as ε=104\varepsilon=10^{-4}, where standard-domain methods fail due to floating-point overflow.

Limitations.

Our implementation has several limitations. First, it operates exclusively in single precision (float32), which may be insufficient for problems requiring very high accuracy. Second, the dense cost matrix storage limits scalability to approximately n=m=16384n=m=16384 on a 16 GB GPU. Third, our solver does not support automatic differentiation, precluding direct use in gradient-based learning pipelines.

Future work.

Several extensions are promising: (1) mixed-precision (fp16/bf16) computation for the Gibbs kernel, with fp32 accumulation for the LogSumExp; (2) sparse or low-rank cost matrix representations to scale beyond n=16384n=16384; (3) a pybind11 wrapper for seamless Python integration; (4) integration with multiscale strategies (Schmitzer, 2019) for further acceleration.

Reproducibility.

All code, experiment scripts, and plotting utilities are publicly available at https://github.com/xiao98/Fast-Sinkhorn-CUDA under the MIT license.

References

  • J. Altschuler, J. Weed, and P. Rigollet (2017) Near-linear time approximation algorithms for optimal transport via Sinkhorn iteration. In Advances in Neural Information Processing Systems (NeurIPS), Vol. 30. Cited by: §2.
  • M. Arjovsky, S. Chintala, and L. Bottou (2017) Wasserstein generative adversarial networks. In Proceedings of the 34th International Conference on Machine Learning (ICML), pp. 214–223. Cited by: §1.
  • B. Charlier, J. Feydy, J. A. Glaunès, F. Collin, and G. Durif (2021) Kernel operations on the GPU, with autodiff, without memory overflows. In Journal of Machine Learning Research, Vol. 22, pp. 1–6. Cited by: §1, §2.
  • L. Chizat, G. Peyré, B. Schmitzer, and F. Vialard (2018) Scaling algorithms for unbalanced optimal transport problems. Mathematics of Computation 87 (314), pp. 2563–2609. Cited by: §2.
  • N. Courty, R. Flamary, D. Tuia, and A. Rakotomamonjy (2017) Optimal transport for domain adaptation. In IEEE Transactions on Pattern Analysis and Machine Intelligence, Vol. 39, pp. 1853–1865. Cited by: §1.
  • M. Cuturi (2013) Sinkhorn distances: lightspeed computation of optimal transport. In Advances in Neural Information Processing Systems (NeurIPS), Vol. 26. Cited by: §1, §2, §3.2.
  • P. Dvurechensky, A. Gasnikov, and A. Kroshnin (2018) Computational optimal transport: complexity by accelerated gradient descent is better than by Sinkhorn’s algorithm. In Proceedings of the 35th International Conference on Machine Learning (ICML), pp. 1367–1376. Cited by: §2.
  • J. Feydy, T. Séjourné, F. Vialard, S. Amari, A. Trouve, and G. Peyré (2019) Interpolating between optimal transport and MMD using Sinkhorn divergences. In Proceedings of the 22nd International Conference on Artificial Intelligence and Statistics (AISTATS), pp. 2681–2690. Cited by: item 3, §1, §2, 2nd item.
  • R. Flamary, N. Courty, A. Gramfort, M. Z. Alaya, A. Boisbunon, S. Chambon, L. Chapel, A. Corenflos, K. Fatras, N. Fournier, L. Gautheron, N. T.H. Gayraud, H. Janati, A. Rakotomamonjy, I. Redko, A. Rolet, A. Schutz, V. Seguy, D. J. Sutherland, R. Tavenard, A. Tong, and T. Vayer (2021) POT: Python optimal transport. Journal of Machine Learning Research 22 (78), pp. 1–8. Cited by: item 3, §2, 1st item.
  • J. Franklin and J. Lorenz (1989) Scaling of matrices to achieve specified row and column sums. Numer. Math. 53, pp. 139–148. Cited by: §A.2.
  • S. Gerber and M. Maggioni (2017) Multiscale strategies for computing optimal transport. In Journal of Machine Learning Research, Vol. 18, pp. 1–32. Cited by: §2.
  • M. J. Kusner, Y. Sun, N. I. Kolkin, and K. Q. Weinberger (2015) From word embeddings to document distances. In Proceedings of the 32nd International Conference on Machine Learning (ICML), pp. 957–966. Cited by: §1.
  • T. Lin, N. Ho, and M. I. Jordan (2019) On efficient optimal transport: an analysis of greedy and accelerated mirror descent algorithms. In Proceedings of the 36th International Conference on Machine Learning (ICML), pp. 3982–3991. Cited by: §2.
  • G. Peyré and M. Cuturi (2019) Computational optimal transport: with applications to data science. Foundations and Trends in Machine Learning 11 (5–6), pp. 355–607. Cited by: §1, §2, §3.1, §3.2, Theorem 1.
  • M. Scetbon, M. Cuturi, and G. Peyré (2021) Low-rank Sinkhorn factorization. In Proceedings of the 38th International Conference on Machine Learning (ICML), pp. 9344–9354. Cited by: §2.
  • B. Schmitzer (2019) Stabilized sparse scaling algorithms for entropy regularized transport problems. SIAM Journal on Scientific Computing 41 (3), pp. A1443–A1481. Cited by: §1, §2, §2, §6.
  • R. Sinkhorn (1967) Diagonal equivalence to matrices with prescribed row and column sums. The American Mathematical Monthly 74 (4), pp. 402–405. Cited by: §2.
  • J. Solomon, F. de Goes, G. Peyré, M. Cuturi, A. Butscher, A. Nguyen, T. Du, and L. Guibas (2015) Convolutional Wasserstein distances: efficient optimal transport on geometric domains. ACM Transactions on Graphics (SIGGRAPH) 34 (4), pp. 66:1–66:11. Cited by: §1.
  • C. Villani (2003) Topics in optimal transportation. Graduate Studies in Mathematics, Vol. 58, American Mathematical Society. Cited by: §1.
  • C. Villani (2009) Optimal transport: old and new. Grundlehren der mathematischen Wissenschaften, Vol. 338, Springer. Cited by: §1.

Appendix A Mathematical Derivation

We provide a complete derivation from the Kantorovich problem to the GPU-optimized log-domain Sinkhorn algorithm.

A.1 From KKT Conditions to Sinkhorn Updates

The Lagrangian of the regularized problem (3) is:

(𝝅,𝜶,𝜷)=C,𝝅+εKL(𝝅𝝁𝝂)𝜶,𝝅𝟏m𝝁𝜷,𝝅𝟏n𝝂.\mathcal{L}(\bm{\pi},\bm{\alpha},\bm{\beta})=\langle C,\bm{\pi}\rangle+\varepsilon\,\mathrm{KL}(\bm{\pi}\|\bm{\mu}\otimes\bm{\nu})-\langle\bm{\alpha},\bm{\pi}\mathbf{1}_{m}-\bm{\mu}\rangle-\langle\bm{\beta},\bm{\pi}^{\top}\mathbf{1}_{n}-\bm{\nu}\rangle. (13)

Setting πij=0\frac{\partial\mathcal{L}}{\partial\pi_{ij}}=0:

Cij+εlogπijμiνjαiβj=0πijε=μiνjexp(αi+βjCijε).C_{ij}+\varepsilon\log\frac{\pi_{ij}}{\mu_{i}\nu_{j}}-\alpha_{i}-\beta_{j}=0\quad\Longrightarrow\quad\pi_{ij}^{\varepsilon}=\mu_{i}\nu_{j}\exp\!\left(\frac{\alpha_{i}+\beta_{j}-C_{ij}}{\varepsilon}\right). (14)

Substituting into the row marginal constraint jπijε=μi\sum_{j}\pi_{ij}^{\varepsilon}=\mu_{i}:

jμiνjexp(αi+βjCijε)\displaystyle\sum_{j}\mu_{i}\nu_{j}\exp\!\left(\frac{\alpha_{i}+\beta_{j}-C_{ij}}{\varepsilon}\right) =μi\displaystyle=\mu_{i} (15)
jνjexp(βjCijε)\displaystyle\sum_{j}\nu_{j}\exp\!\left(\frac{\beta_{j}-C_{ij}}{\varepsilon}\right) =exp(αiε)\displaystyle=\exp\!\left(\frac{-\alpha_{i}}{\varepsilon}\right) (16)
αi\displaystyle\alpha_{i} =εlogjνjexp(βjCijε).\displaystyle=-\varepsilon\log\sum_{j}\nu_{j}\exp\!\left(\frac{\beta_{j}-C_{ij}}{\varepsilon}\right). (17)

The column constraint yields the analogous update for βj\beta_{j}.

A.2 Convergence Rate

Proof sketch of Theorem 1.

Define the operator T:(𝜶,𝜷)(𝜶,𝜷)T:(\bm{\alpha},\bm{\beta})\mapsto(\bm{\alpha}^{\prime},\bm{\beta}^{\prime}) by the Sinkhorn updates. The operator TT is a contraction in the Hilbert projective metric on the positive cone. The contraction rate is determined by the Birkhoff contraction coefficient:

λ=tanh(R4ε)2e2R/ε,\lambda=\tanh\!\left(\frac{R}{4\varepsilon}\right)^{2}\leq e^{-2R/\varepsilon}, (18)

where R=maxi,jCijmini,jCijR=\max_{i,j}C_{ij}-\min_{i,j}C_{ij}. This follows from classical results on positive matrix scaling [Franklin and Lorenz, 1989]. ∎

A.3 LogSumExp Numerical Analysis

The standard computation logjexp(xj)\log\sum_{j}\exp(x_{j}) fails when maxjxj>88\max_{j}x_{j}>88 (overflow of exp\exp in float32) or when maxjxj<88\max_{j}x_{j}<-88 and all terms underflow to zero (yielding log(0)=\log(0)=-\infty). In our setting, xj=(βjCij)/ε+logνjx_{j}=(\beta_{j}-C_{ij})/\varepsilon+\log\nu_{j}, and for ε=0.01\varepsilon=0.01 with Cij[0,1]C_{ij}\in[0,1], we have xj[100/ε,0][104,0]x_{j}\in[-100/\varepsilon,0]\approx[-10^{4},0]—clearly problematic.

The stabilized version M+logjexp(xjM)M+\log\sum_{j}\exp(x_{j}-M) with M=maxjxjM=\max_{j}x_{j} ensures:

  • The largest exponent is e0=1e^{0}=1 (no overflow).

  • All exponents satisfy exjM(0,1]e^{x_{j}-M}\in(0,1] (no overflow).

  • The sum is at least 1 (from the jj^{*} achieving the max), so log()0\log(\cdot)\geq 0 (stable).

Appendix B Extended Benchmark Results

Table 3: Extended benchmark results across all tested configurations.
nn ε\varepsilon Time (ms) Iterations Transport cost Converged
256 0.1 0.15 31 0.0413
256 0.01 0.32 97 0.0395
256 0.001 0.65 340 0.0391
512 0.1 0.64 30 0.0391
512 0.01 1.26 80 0.0366
512 0.001 2.41 284 0.0406
1024 0.1 2.46 27 0.0429
1024 0.01 5.09 100 0.0372
1024 0.001 9.99 319 0.0377
2048 0.1 10.88 29 0.0394
2048 0.01 20.66 118 0.0400
2048 0.001 40.26 342 0.0376
4096 0.1 44.68 25 0.0373
4096 0.01 89.09 107 0.0403
4096 0.001 175.00 306 0.0370
8192 0.1 176.51 30 0.0421
8192 0.01 371.60 82 0.0406
8192 0.001 714.88 294 0.0412

Appendix C GPU Specifications

Table 4: GPU hardware specifications.
Property Value
Device NVIDIA GeForce RTX 3090
Compute capability 8.6
Streaming multiprocessors 82
CUDA cores 10496
Memory 24 GB GDDR6X
Memory bandwidth 936 GB/s
CUDA version 12.2
Baseline GPU (Python baselines)
Device NVIDIA Tesla T4
Compute capability 7.5
Streaming multiprocessors 40
CUDA cores 2560
Memory 16 GB GDDR6
Memory bandwidth 320 GB/s

Appendix D Core Kernel Implementation

For reference, we include the complete CUDA kernel for the 𝜶\bm{\alpha}-update (the 𝜷\bm{\beta}-update is analogous):

__global__ void updateAlphaKernel(
    const float* C, const float* log_nu,
    const float* beta, float* alpha,
    int n, int m, float eps)
{
    const int i = blockIdx.x;
    if (i >= n) return;
    const float inv_eps = 1.0f / eps;
    const float* Ci = C + i * m;

    // Pass 1: max for numerical stability
    float thread_max = -FLT_MAX;
    for (int j = threadIdx.x; j < m; j += blockDim.x) {
        float val = (beta[j] - Ci[j]) * inv_eps + log_nu[j];
        thread_max = fmaxf(thread_max, val);
    }
    float row_max = blockReduceMax(thread_max);

    __shared__ float s_max;
    if (threadIdx.x == 0) s_max = row_max;
    __syncthreads();
    row_max = s_max;

    // Pass 2: stable sum
    float thread_sum = 0.0f;
    for (int j = threadIdx.x; j < m; j += blockDim.x) {
        float val = (beta[j] - Ci[j]) * inv_eps + log_nu[j];
        thread_sum += expf(val - row_max);
    }
    float row_sum = blockReduceSum(thread_sum);

    if (threadIdx.x == 0) {
        alpha[i] = -eps * (row_max
                   + logf(fmaxf(row_sum, 1e-30f)));
    }
}